/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */


#ifdef SYNCCHECK

#include "SyncChecker.h"


unsigned CSyncChecker::g_checksum;
int CSyncChecker::inSyncedCode;


#endif // SYNCDEBUG
