/*
 * SpringDebug.cpp
 *
 *  Created on: Nov 8, 2019
 *      Author: rlcevg
 */

#include "spring/SpringDebug.h"

namespace circuit {

CDebug::CDebug()
{
}

CDebug::~CDebug()
{
}

} // namespace circuit
