/*
 * UnitManager.cpp
 *
 *  Created on: Jan 15, 2015
 *      Author: rlcevg
 */

#include "unit/UnitManager.h"

namespace circuit {

IUnitManager::IUnitManager()
{
}

IUnitManager::~IUnitManager()
{
}

} // namespace circuit
