/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#include "NullSound.h"
#include "NullAudioChannel.h"

#include "System/Sound/SoundLog.h"

void NullSound::PrintDebugInfo() {
	LOG("Null Sound System");
}

