/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#ifndef WEAPON_PROJECTILE_TYPES_H
#define WEAPON_PROJECTILE_TYPES_H

enum {
	WEAPON_BASE_PROJECTILE           = (1 <<  0),
	WEAPON_BEAMLASER_PROJECTILE      = (1 <<  1),
	WEAPON_EMG_PROJECTILE            = (1 <<  2),
	WEAPON_EXPLOSIVE_PROJECTILE      = (1 <<  3),
	WEAPON_FIREBALL_PROJECTILE       = (1 <<  4),
	WEAPON_FLAME_PROJECTILE          = (1 <<  5),
	WEAPON_LARGEBEAMLASER_PROJECTILE = (1 <<  6),
	WEAPON_LASER_PROJECTILE          = (1 <<  7),
	WEAPON_LIGHTNING_PROJECTILE      = (1 <<  8),
	WEAPON_MISSILE_PROJECTILE        = (1 <<  9),
	WEAPON_STARBURST_PROJECTILE      = (1 << 10),
	WEAPON_TORPEDO_PROJECTILE        = (1 << 11),
};

#endif

