#ifndef INC_EXTERNAI_HDR
#define INC_EXTERNAI_HDR

// Spring AI Interface Headers
#include "LegacyCpp/aibase.h"            // DLL exports and definitions
#include "LegacyCpp/IGlobalAI.h"         // Main AI file
#include "LegacyCpp/IAICallback.h"       // Callback functions
#include "LegacyCpp/IGlobalAICallback.h" // AI Interface
#include "LegacyCpp/IAICheats.h"         // AI Cheat Interface
#include "LegacyCpp/UnitDef.h"           // AI-side UnitDef structure definition
#include "LegacyCpp/CommandQueue.h"      // AI-side Unit Command Queues
#include "LegacyCpp/FeatureDef.h"        // AI-side FeatureDef structure definition
#include "LegacyCpp/MoveData.h"          // AI-side Types of Movement units can have
#include "LegacyCpp/WeaponDef.h"         // AI-side WeaponDef structure definition

#include "ExternalAI/Interface/aidefines.h"          // SNPRINTF, STRCPY, ...

using namespace springLegacyAI;

#endif
