/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#include "Rectangle.h"

CR_BIND(SRectangle, )
CR_REG_METADATA(SRectangle, (
	CR_MEMBER(x1),
	CR_MEMBER(y1),
	CR_MEMBER(x2),
	CR_MEMBER(y2)
))
