/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#ifndef _CREG_RUNTIME_TESTS_H
#define _CREG_RUNTIME_TESTS_H

namespace creg {
	bool RuntimeTest();
}

#endif // _CREG_RUNTIME_TESTS_H

