/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#ifndef _CPPTESTAI_CPPTESTAI_H
#define _CPPTESTAI_CPPTESTAI_H

// generated by the C++ Wrapper scripts
#include "OOAICallback.h"

namespace cpptestai {

/**
 * This is the main C++ entry point of this AI.
 * 
 * @author	Robin Vobruba <hoijui.quaero@gmail.com>
 */
class CCppTestAI {

private:
	springai::OOAICallback* callback;
	int skirmishAIId;

public:
	CCppTestAI(springai::OOAICallback* callback);
	~CCppTestAI();

	int HandleEvent(int topic, const void* data);
}; // class CCppTestAI

} // namespace cpptestai

#endif // _CPPTESTAI_CPPTESTAI_H

