/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#include "AudioChannel.h"
#include "System/Sound/ISoundChannels.h"

IAudioChannel* Channels::BGMusic = NULL;
IAudioChannel* Channels::General = NULL;
IAudioChannel* Channels::Battle = NULL;
IAudioChannel* Channels::UnitReply = NULL;
IAudioChannel* Channels::UserInterface = NULL;
