/*
 * ResurrectTask.h
 *
 *  Created on: Apr 16, 2020
 *      Author: rlcevg
 */

#ifndef SRC_CIRCUIT_TASK_BUILDER_RESURRECTTASK_H_
#define SRC_CIRCUIT_TASK_BUILDER_RESURRECTTASK_H_

namespace circuit {

class CResurrectTask {
public:
	CResurrectTask();
	virtual ~CResurrectTask();
	// TODO: similar to ReclaimTask?
};

} // namespace circuit

#endif // SRC_CIRCUIT_TASK_BUILDER_RESURRECTTASK_H_
