/*
 * SpringDebug.h
 *
 *  Created on: Nov 8, 2019
 *      Author: rlcevg
 */

#ifndef SRC_CIRCUIT_SPRING_SPRINGDEBUG_H_
#define SRC_CIRCUIT_SPRING_SPRINGDEBUG_H_

namespace circuit {

class CDebug {
public:
	CDebug();
	virtual ~CDebug();
};

} // namespace circuit

#endif // SRC_CIRCUIT_SPRING_SPRINGDEBUG_H_
