/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#ifndef DEBUGCOLVOLDRAWER_H_
#define DEBUGCOLVOLDRAWER_H_


namespace DebugColVolDrawer
{
	extern bool enable;

	void Init();
	void Kill();
	void Draw();
}


#endif /* DEBUGCOLVOLDRAWER_H_ */
