/* This file is part of the Spring engine (GPL v2 or later), see LICENSE.html */

#ifndef _AL_SHARED_H_
#define _AL_SHARED_H_

#include <al.h>

bool CheckError(const char* msg);

#endif // _AL_SHARED_H_
