/*
 * ResurrectTask.cpp
 *
 *  Created on: Apr 16, 2020
 *      Author: rlcevg
 */

#include "task/builder/ResurrectTask.h"

namespace circuit {

CResurrectTask::CResurrectTask()
{
}

CResurrectTask::~CResurrectTask()
{
}

} // namespace circuit
